/*
* C_lib Copyright (C) 2021 Wael Karram.
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, version 3 of the License.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

/* This header defines a single-linked-list and related functions. */
/* Use the single nodes. */
#include <stdlib.h> /* Used in allocations and deallocations. */
#ifndef SINGLE_LINKED_LIST_H
#define SINGLE_LINKED_LIST_H
#include "../nodes/single_node.h"
/* Include assert.h to allow code generation, but only run with debug flag. */
#ifdef DEBUG
#ifndef NDEBUG
#define NDEBUG
#endif /* NDEBUG. */
#endif /* DEBUG. */
#include <assert.h>

/* Constants. */
/* Function error/return code on success. */
#define SUCCESS_CODE_SINGLE_LINKED_LIST_C			1
/* Function error/return code on failure. */
#define FAILURE_CODE_SINGLE_LINKED_LIST_C			0
/* Function error/return code on failure to find index. */
#define INDEX_NOT_FOUND_CODE_SINGLE_LINKED_LIST_C		-1
/* Function error/return code on invalid input. */
#define INVALID_INPUT_CODE_SINGLE_LINKED_LIST_C			-2
/* Error code returned when input fails sanity check. */
#define CODE_FREE_SUCCESS_MALFORMED_SINGLE_LINKED_LIST_C	-1
/* The default index of the list's head. */
#define INDEX_HEAD_SINGLE_LINKED_LIST_C				0

/* The linked list's struct. */
typedef struct linked_list {
	/* Stores the list's length. */
	size_t length;
	/* Stores the head node pointer. */
	node_t *head;
} linked_list_t;

/* This function initializes a new linked list. */
static inline linked_list_t* initialize_single_linked_list();
/* This function appends a node to the given linked list. */
static inline void append_node_single_linked_list(linked_list_t *list, node_t *node);
/* This function prepends a node to the given linked list. */
static inline void prepend_node_single_linked_list(linked_list_t *list, node_t *node);
/* This function creates and appends a node to the given linked list. */
static inline int append_node_data_single_linked_list(linked_list_t *list, void *data);
/* This function creates and prepends a node to the given linked list. */
static inline int prepend_node_data_single_linked_list(linked_list_t *list, void *data);
/* This function inserts a node at the ith place. */
static inline void add_node_i_single_linked_list(linked_list_t *list, node_t *node, size_t i);
/* This function creates and inserts a node at the ith place. */
static inline int add_node_data_i_single_linked_list(linked_list_t *list, void *data, size_t i);
/* This function reads the data at the ith node. */
static inline void* read_node_i_single_linked_list(linked_list_t *list, size_t i);
/* This function retrieves a pointer to the ith node. */
static inline node_t* get_node_i_single_linked_list(linked_list_t *list, size_t i);
/* This function attempts to find the index of a given node, by pointer. */
static inline size_t get_node_index_pointer_single_linked_list(linked_list_t *list, node_t *node);
/* This function attempts to find the index of a given node, by data. */
static inline size_t get_node_index_data_single_linked_list(linked_list_t *list, void *data, int (*comparison_function)(const void*, const void*));
/* This function deletes the ith node. */
static inline void remove_node_i_single_linked_list(linked_list_t *list, size_t i);
/* This function deletes the ith node and the data stored within. */
static inline void remove_node_i_data_single_linked_list(linked_list_t *list, size_t i, void (*free_function)(const void*));
/* This function reads the data at the first node. */
static inline void* read_node_head_single_linked_list(linked_list_t *list);
/* This function retrieves the first node. */
static inline node_t* get_node_head_single_linked_list(linked_list_t *list);
/* This function deletes the first node. */
static inline void remove_node_head_single_linked_list(linked_list_t *list);
/* This function deletes the first node and the data thereof. */
static inline void remove_node_head_data_single_linked_list(linked_list_t *list, void (*free_function)(const void*));
/* This function reads the data at the last node. */
static inline void* read_node_tail_single_linked_list(linked_list_t *list);
/* This function retrieves the last node. */
static inline node_t* get_node_tail_single_linked_list(linked_list_t *list);
/* This function deletes the last node. */
static inline void remove_node_tail_single_linked_list(linked_list_t *list);
/* This function deletes the last node and the data thereof. */
static inline void remove_node_tail_data_single_linked_list(linked_list_t *list, void (*free_function)(const void*));
/* This function deletes all the nodes of the given list and the list itself. */
static inline int clear_list_single_linked_list(linked_list_t **list);
/* This function deletes all the nodes and data stored within of the given list and the list itself. */
static inline int clear_list_data_single_linked_list(linked_list_t **list, void (*free_function)(const void*));
/* This function deletes all the nodes of a linked list. */
static inline void clear_list_nodes_single_linked_list(linked_list_t *list);
/* This function deletes all the nodes of a linked list and the data thereof. */
static inline void clear_list_nodes_data_single_linked_list(linked_list_t *list, void (*free_function)(const void*));
/* This function attempts to serialize a linked list into an array, it relies on an array setting function and an allocated buffer. */
static inline void *serialize_linked_list(linked_list_t *list, void (*array_setting_function)(void*, size_t, void*), void *array);
#endif /* SINGLE_LINKED_LIST_H */
