/* Include statements. */
#include <stdio.h>
#include <assert.h>
#include "../linked_lists/queue.c"

/* Define constants. */
#define PRINT_INFO_TESTING_START		"Starting tests.\n"
#define PRINT_INFO_TESTING_END			"Done testing.\n"
#define ERROR_TEXT_RANDOM_QUEUE_FAIL_ALLOCATION	"Long queue test failed because of allocation errors!\n"
#define ERROR_SERIALIZATION_TEST_FAILURE	"Failed the serialization test.\n"
#define SOME_TEST_FAILED			"Recheck the tests, one of them failed!\n"
#define INITIAL_QUEUE_LENGTH			0
#define QUEUE_LENGTH_ONE			1
#define QUEUE_LENGTH_LONG_TEST			4096
#define SERIALIZATION_TEST_QUEUE_LENGTH		5

/* Define functions. */
node_t* generate_random_node();
queue_t* generate_long_random_queue(int length);
void set_array_value_index(void *array, size_t index, void *value);
int queue_count_length(queue_t *queue);
int test_serialization();
int test_queues();

/* Starting point. */
int main() {
	printf(PRINT_INFO_TESTING_START);
	if (test_queues())
		printf(SOME_TEST_FAILED);
	printf(PRINT_INFO_TESTING_END);
}

/* This function generates a random number and puts it in a newly allocated node. */
/* Caller has to free memory afterwards, returns NULL on failure. */
node_t* generate_random_node() {
	/* The node to return. */
	node_t* result;
	int *value;

	/* Attempt to allocate the node. */
	result = (node_t*) malloc(sizeof(node_t));
	if (result == NULL)
		return NULL;
	value = (int*) malloc(sizeof(int));
	if (value == NULL) {
		free(result);
		return NULL;
	}
	*value = random();

	/* Store and return. */
	result->next = NULL;
	result->data = value;
	return result;
}

/* This function generates a long queue with random values in the nodes, returns NULL on failures. */
/* Length denotes the length of the queue. */
queue_t* generate_long_random_queue(int length) {
	/* Sanity check. */
	if (length <= 0)
		return NULL;
	/* Variables. */
	queue_t *queue;
	node_t *current;
	int i;

	/* Allocate the queue. */
	queue = initialize_queue();
	if (queue == NULL)
		return NULL;

	/* Fill the queue with random nodes. */
	/* First node. */
	current = generate_random_node();
	if (current == NULL) {
		free(queue);
		return NULL;
	}
	/* Attempt to enqueue. */
	enqueue(queue, current);

	/* Loop and enqueue the rest. */
	for (i = 1; i < length; i++) {
		/* Generate and attempt to enqueue. */
		current = generate_random_node();
		if (current == NULL) {
			/* Abort. */
			free(current->data);
			free(current);
			clear_queue_data(&queue, NULL);
			return NULL;
		}
		enqueue(queue, current);
	}

	/* Return the result. */
	return queue;
}

/* This function is used to set a value at an index in the array. */
void set_array_value_index(void *array, size_t index, void *value) {
	/* Cast. */
	int *arr = (int*)array;
	int *val = (int*)value;
	/* Set the value. */
	arr[index] = *val;
}

/* This function checks the actual length of a queue. */
int queue_count_length(queue_t *queue) {
	if (!queue || !queue->list || !queue->list->head)
		return 0;
	/* Loop and count. */
	node_t *current = queue->list->head;
	int count = 0;
	while (current) {
		count++;
		current = current->next;
	}
	/* Return the result. */
	return count;
}

/* This function attempts to allocate a queue, fill it, then serialize it. */
int test_serialization() {
	/* Local variables. */
	int i;
	size_t length;
	/* Used to allocate and store a value. */
	int *value, *array;

	/* Allocations. */
	/* Initialize the queue. */
	queue_t *queue = initialize_queue();
	if (!queue)
		return 1;

	/* Allocate the array. */
	array = (int*) malloc(sizeof(int) * SERIALIZATION_TEST_QUEUE_LENGTH);
	if (!array) {
		free(queue);
		return 1;
	}

	/* Fill the queue. */
	for (i = 0; i < SERIALIZATION_TEST_QUEUE_LENGTH; i++) {
		/* Allocate the value store. */
		value = (int*) malloc(sizeof(int));
		if (!value) {
			/* Clear the queue. */
			clear_queue_data(&queue, NULL);
			/* Free the array. */
			free(array);
			/* Failed the test, technical reasons. */
			return 1;
		}
		/* Set the value. */
		*value = i;
		/* Enqueue the data. */
		enqueue_data(queue, value);
	}

	/* Serialize the queue and check if the serialization failed. */
	if (!serialize_queue(queue, set_array_value_index, array)) {
		/* Clear the queue. */
		clear_queue_data(&queue, NULL);
		/* Free the array. */
		free(array);
		/* Failed the test. */
		return 0;
	}

	/* Check the values. */
	for (i = 0; i < SERIALIZATION_TEST_QUEUE_LENGTH ; i++) {
		if (array[i] != (SERIALIZATION_TEST_QUEUE_LENGTH - 1 - i)) {
			/* Clear the queue. */
			clear_queue_data(&queue, NULL);
			/* Free the array. */
			free(array);
			/* Failed the test. */
			return 0;
		}
	}

	/* Passed the tests. */
	/* Clear the queue. */
	clear_queue_data(&queue, NULL);
	/* Free the array. */
	free(array);

	return 1;
}

/* This function does the testing. */
int test_queues() {
	/* Local variables. */
	queue_t *queue;
	node_t *node, *tail;

	/* Initialize an empty queue. */
	queue = initialize_queue();
	/* Check that it was correctly created. */
	assert(queue);
	assert(queue->list);
	assert(queue->list->length == INITIAL_QUEUE_LENGTH);
	assert(!(queue->list->head));
	assert(!(queue->tail));

	/* Generate and store the node. */
	node = generate_random_node();

	/* Attempt to add it to the queue. */
	enqueue(queue, node);
	/* Check that the queue grew by one and the head got set. */
	assert(queue->list->length == QUEUE_LENGTH_ONE);
	assert(queue->list->head && queue->list->head == node);
	assert(queue->tail && queue->tail == node);
	assert(get_node_i_single_linked_list(queue->list, 0) == node);
	assert(get_node_index_pointer_single_linked_list(queue->list, node) == 0);

	/* Attempt to free the nodes and then the queue. */
	node = dequeue(queue);
	assert(queue->list->length == INITIAL_QUEUE_LENGTH);
	assert(!(queue->list->head));
	assert(!(queue->tail));
	free(node->data);
	free(node);
	node = NULL;
	free(queue->list);
	free(queue);
	queue = NULL;

	/* Attempt to allocate a long queue. */
	queue = generate_long_random_queue(QUEUE_LENGTH_LONG_TEST);
	if (queue == NULL) {
		printf(ERROR_TEXT_RANDOM_QUEUE_FAIL_ALLOCATION);
		return 1;
	}
	assert(queue->list->length == QUEUE_LENGTH_LONG_TEST);
	assert(queue->list->length == queue_count_length(queue));
	/* Test the freeing functions. */
	clear_queue_data(&queue, NULL);
	assert(!queue);

	/* Serialization test. */
	if (!test_serialization()) {
		printf(ERROR_SERIALIZATION_TEST_FAILURE);
		return 1;
	}

	/* Ran successfully, return 0. */
	return 0;
}
