/*
* C_lib Copyright (C) 2021 Wael Karram.
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, version 3 of the License.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

/* This file defines a simple binary tree and related functions thereof. */
#ifndef RED_BLACK_TREE_H
#define RED_BLACK_TREE_H
#include "../../nodes/rb_tree_node.h"
#include <math.h>
#include <stdlib.h>
#include <errno.h>
/* Include assert.h to allow code generation, but only run with debug flag. */
#ifndef DEBUG
#ifndef NDEBUG
#define NDEBUG
#endif /* NDEBUG. */
#endif /* DEBUG. */
#include <assert.h>

/* Define constants. */
#define TRUE							1
#define FALSE							0
/* Return code for success. */
#define RED_BLACK_TREE_SUCCESS_CODE				0
/* Return code for failure. */
#define RED_BLACK_TREE_FAILURE_CODE				1
/* Return code for comparison, smaller. */
#define RED_BLACK_TREE_COMPARISON_SMALLER_THAN			-1
/* Return code for comparison, equal. */
#define RED_BLACK_TREE_COMPARISON_EQUAL_TO			0
/* Return code for comparison, larger. */
#define RED_BLACK_TREE_COMPARISON_LARGER_THAN			1

/* Define tree struct. */
typedef struct rb_tree_t {
	/* Root node. */
	rb_tree_node_t *root;
} rb_tree_t;

/* Enum to define right an left sides of the root. */
enum rb_child_side { LEFT, RIGHT };
typedef enum rb_child_side rb_child_side;

/* Macro to get the side of the child in relation to the parent thereof. */
#define CHILD_SIDE(N) (N == (N->parent)->right ? RIGHT : LEFT)

/* Definitions of macros for rotation functions. */
#define rotate_direction(T, N, direction)	rotate_direction_root(T, N, direction)
#define rotate_left(T, N)			rotate_direction_root(T, N, LEFT)
#define rotate_right(T, N)			rotate_direction_root(T, N, RIGHT)

/* Function definitions. */
/* This function initializes an empty red-black tree. */
static inline rb_tree_t* initialize_rb_tree();
/* This function initializes a red-black tree and stores a value in the root. */
static inline rb_tree_t* initialize_rb_tree_root_store(const void* const data);
/* This function attempts to find the parent of a new node. */
static rb_tree_t* find_parent(const void* data, int (*comparison_function)(const void*, const void*));
/* This function inserts a new node. */
static rb_tree_t* insert_node(const void* data, int (*comparison_function)(const void*, const void*));
/* This function frees a whole Red-Black tree, not including data. */
static inline void free_rb_tree(rb_tree_t* tree);
/* This function frees a whole Red-Black tree, including data. */
static inline void free_rb_tree_data(rb_tree_t* tree, void (*free_function)(const void*));
/* This function handles removal of a node. */
static inline void remove_rb_tree_node(rb_tree_t* tree, rb_tree_node_t* node);
/* This function handles removal of a node and the data thereof. */
static inline void remove_rb_tree_node_data(rb_tree_t* tree, rb_tree_node_t* node, void (*free_function)(const void*));
/* This function finds a node by data search. */
static inline rb_tree_node_t* find_rb_tree_node_data(const rb_tree_t* const tree, const void* data, int (*comparison_function)(const void*, const void*));

/* Helper functions. */
/* This function recursively frees a red-black tree from the given node downwards, doesn't touch the stored data. */
static void free_rb_tree_recursive(rb_tree_node_t* root);
/* This function recursively frees a red-black tree from the given node downwards, including stored data. */
static void free_rb_tree_data_recursive(rb_tree_node_t* root, void (*free_function)(const void*));
/* This function recursively searches for a node by data. */
static rb_tree_node_t* find_rb_tree_node_data_recursive(rb_tree_node_t* root, const void* data, int (*comparison_function)(const void*, const void*));
/* This function applies a rotation upon a tree at the given root. */
static inline rb_tree_node_t* rotate_direction_root(rb_tree_t* tree, rb_tree_node_t* root, rb_child_side side);
/* Insert, internal implementation. */
static inline void rb_tree_insert_implementation(rb_tree_t* tree, rb_tree_node_t* node, rb_tree_node_t* parent, rb_child_side side);

#endif /* RED_BLACK_TREE_H */
