/*
* C_lib Copyright (C) 2021 Wael Karram.
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, version 3 of the License.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

/* This file contains the function implementations for binary search trees. */
/* Include statements. */
/* Include guard. */
#ifndef BINARY_TREE_C_INCLUDE
#define BINARY_TREE_C_INCLUDE
#include "../binary_tree/binary_tree.c"
#endif /* BINARY_TREE_C_INCLUDE */
#include "binary_search_tree.h"

/* This function initializes a new empty tree node, will return NULL on failure. */
static inline tree_node_t* initialize_binary_search_tree_node() {
	/* Delegate internally. */
	return initialize_tree_node();
}

/* This function initializes a new tree node and attempts to store the given data within, will return NULL on failure. */
static inline tree_node_t* initialize_binary_search_tree_node_store(void* const data) {
	/* Delegate internally. */
	return initialize_tree_node_store(data);
}

/* This function is used to recursively fill an array into a BST. */
static inline tree_node_t* initialize_binary_search_tree_recursive(void** const array, size_t index_start, size_t index_end) {
	/* Sanity check. */
	assert(array);

	/* Local variables. */
	size_t middle_index;
	tree_node_t *root;

	/* Calculate the middle index. */
	middle_index = index_start + index_end;
	middle_index /= ((size_t)2L);

	/* Allocate the node. */
	root = (tree_node_t*) malloc(sizeof(tree_node_t));
	if (!root)
		return NULL;

	/* Set the value. */
	root->data = array[middle_index];

	/* Check if there is a need to recurse. */
	if (index_start == index_end) {
		/* Make a leaf and stop. */
		root->right = NULL;
		root->left = NULL;
		return root;
	}

	/* Recursive case. */
	root->right = initialize_binary_search_tree_recursive(array, middle_index, index_end);
	root->left = initialize_binary_search_tree_recursive(array, index_start, middle_index);

	/* Check if allocation failed in any of the recursive cases. */
	if (!(root->right) || !(root->left)) {
		free(root->right);
		free(root->left);
		return NULL;
	}

	/* Return the node. */
	return root;
}

/* This function creates a BST for a sorted array. */
static inline tree_node_t* initialize_binary_search_tree(void** const array, size_t length) {
	/* Sanity check. */
	assert(array && length);

	/* Delegate. */
	return initialize_binary_search_tree_recursive(array, 0, length - 1);
}

/* This function links a node to the right, will overwrite links!. */
static inline void link_binary_search_tree_node_right(tree_node_t *parent, tree_node_t* const child) {
	/* Delegate. */
	link_tree_node_right(parent, child);
}

/* This function links a node to the left, will overwrite links!. */
static inline void link_binary_search_tree_node_left(tree_node_t *parent, tree_node_t* const child) {
	/* Delegate. */
	link_tree_node_left(parent, child);
}

/* This function attempts to free a whole tree (recursively), while not touching the data. */
static inline void free_binary_search_tree(tree_node_t *root) {
	/* Delegate. */
	free_tree(root);
}

/* This function attempts to free a whole tree (recursively), while using free_function to free the data, if the latter is NULL - will use stdlib free. */
static inline void free_binary_search_tree_data(tree_node_t *root, void (*free_function)(const void*)) {
	/* Delegate. */
	free_tree_data(tree, free_function);
}

/* Note that the implementation disallows duplication! */
/* This function inserts data into the tree, will silently fail on invalid input. Comparison function should return 0 for equal, 1 for larger than and -1 for less than. */
static inline void insert_node_into_binary_search_tree(tree_node_t *root, tree_node_t* const to_insert, int (*comparison_function)(const void*, const void*)) {
	/* Local variables. */
	int comparison_result;

	/* Sanity check. */
	assert(root && comparison_function);

	/* Compare. */
	comparison_result = (*comparison_function)(root->data, to_insert->data);
	/* Check if the value is already there. */
	if (comparison_result == TREE_COMPARISON_EQUAL_TO)
		return;
	/* Check if the root is smaller. */
	if (comparison_result == TREE_COMPARISON_SMALLER_THAN) {
		/* The root is smaller. */
		if (!(root->right)) {
			/* Insert to the right. */
			root->right = to_insert;
			return;
		}
		/* Recurse. */
		return insert_node_into_binary_search_tree(root->right, to_insert, comparison_function);
	} else {
		/* The root is larger. */
		if (!(root->left)) {
			/* Insert to the left. */
			root->right = to_insert;
			return;
		}
		/* Recurse. */
		return insert_node_into_binary_search_tree(root->left, to_insert, comparison_function);
	}
	/* Shouldn't ever reach here! */
}

/* This function searches for data in the tree. */
/* This function inserts data into the tree, will fail on invalid input. Comparison function should return 0 for equal, 1 for larger than and -1 for less than. */
static inline tree_node_t* search_in_binary_search_tree(const tree_node_t* const root, const void* const data, int (*comparison_function)(const void*, const void*)) {
	/* Local variables. */
	int comparison_result;

	/* Sanity check. */
	assert(root && comparison_function);

	/* Check if we found the right node, and recurse accordingly. */
	comparison_result = (*comparison_function)(root->data, data);
	return ((comparison_result == 0) ? root : ((comparison_result == 1) ? search_in_binary_search_tree(root->left, data, comparison_function) : search_in_binary_search_tree(root->right, data, comparison_function)));
}

/* This function finds the tree's depth (at the deepest leaf). */
static inline size_t find_binary_search_tree_depth(const tree_node_t* const root) {
	/* Delegate. */
	return find_tree_depth(root);
}

/* This function fills an allocated array in a sorted manner with the values from a BST. */
/* Takes the tree's root, the array to fill, a pointer to the current fill index, and a value of the array's maximal length. */
/* Doesn't do much sanity checking on the input, returns defined constants on failure/success. */
static inline void fill_array_binary_search_tree(tree_node_t *root, void **array, size_t *index, const size_t length) {
	/* Sanity check. */
	assert(root && array);

	/* Stopping condition. */
	if (*index >= length)
		return;

	/* Check if we have reached a leaf. */
	if (!(root->left) && !(root->right)) {
		/* Insert into the array, and advance. */
		array[*index] = root->data;
		(*index)++;
		return;
	}

	/* Non-leaf node. */
	/* Recurse to the left, add the current node's data, and then recurse to the right. */
	/* Left. */
	if (root->left)
		fill_array_binary_search_tree(root->left, array, index, length);
	/* Check the length. */
	if (*index >= length)
		return;
	/* Current. */
	array[*index] = root->data;
	(*index)++;
	/* Check the length. */
	if (*index >= length)
		return;
	/* Right. */
	if (root->right != NULL)
		fill_array_binary_search_tree(root->right, array, index, length);
}

/* This function creates a sorted array, from a binary search tree. */
/* Sets the resize status accodrdingly (in case it failed or not). */
/* Returns a pointer to the first cell in the array on success, NULL on failure. */
static inline void **sorted_array_from_binary_search_tree(tree_node_t *root, size_t *length, int *resize_status) {
	/* Sanity check. */
	assert(root && length);

	/* Local variables. */
	size_t height, array_size, index;
	void **array;

	/* Set resize status to true by default. */
	*resize_status = TRUE;

	/* Find the tree's height. */
	height = find_binary_search_tree_depth(root);

	/* Derive the array's (maximal) size based on said height (assumes a balanced and near-full tree). */
	array_size = (size_t) (pow(2, height + 1) - 1);

	/* Attempt to allocate the array. */
	array = (void**) calloc(array_size, sizeof(void*));
	/* Check for allocation failures. */
	if (!array)
		return NULL;

	/* Try to fill. */
	index = 0;
	/* Delegate. */
	fill_array_binary_search_tree(root, array, &index, array_size);
	/* Set the array's length. */
	*length = index;
	/* Resize the array if needed. */
	if (index < array_size) {
		/* Check for (non-fatal) failure. */
		if (!reallocarray(array, index, sizeof(void*)) && errno == ENOMEM)
			*resize_status = FALSE;
		else
			*resize_status = TRUE;
	}

	/* Success. */
	return array;
}
