/*
* C_lib Copyright (C) 2021 Wael Karram.
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, version 3 of the License.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

/* This file contains the function implementations for binary trees. */
#include "binary_tree.h"

/* This function initializes a new empty tree node, will return NULL on failure. */
static inline tree_node_t* initialize_tree_node() {
	/* Attempt to allocate and return the result. */
	tree_node_t *result = (tree_node_t*) malloc(sizeof(tree_node_t));
	if (!result)
		return NULL;
	result->right = NULL;
	result->left = NULL;
	result->data = NULL;
	return result;
}

/* This function initializes a new tree node and attempts to store the given data within, will return NULL on failure. */
static inline tree_node_t* initialize_tree_node_store(void *data) {
	/* Attempt to allocate. */
	tree_node_t *result = (tree_node_t*) malloc(sizeof(tree_node_t));
	/* Check, fill and return. */
	if (!result)
		return NULL;
	result->right = NULL;
	result->left = NULL;
	result->data = data;
	return result;
}

/* This function links a node to the right, will overwrite links! */
static inline void link_tree_node_right(tree_node_t *parent, tree_node_t* const child) {
	/* Sanity check. */
	assert(parent && child);

	/* Link. */
	parent->right = child;
}

/* This function links a node to the left, will overwrite links! */
static inline void link_tree_node_left(tree_node_t *parent, tree_node_t* const child) {
	/* Sanity check. */
	assert(parent && child);

	/* Link. */
	parent->left = child;
}

/* This function attempts to free a whole tree (recursively), while not touching the data. */
static inline void free_tree(tree_node_t *root) {
	/* Stopping conditions. */
	if (!root)
		return;
	if (!(root->right) && !(root->left)) {
		free(root);
		return;
	}
	/* Recursive case. */
	if (root->right) {
		free_tree(root->right);
		root->right = NULL;
	}
	if (root->left) {
		free_tree(root->left);
		root->left = NULL;
	}
	/* Free the current node. */
	free(root);
}

/* This function attempts to free a whole tree (recursively), while using free_function to free the data, if the latter is NULL - will use stdlib free. */
static inline void free_tree_data(tree_node_t *root, void (*free_function)(const void*)) {
	/* Stopping conditions. */
	if (!root)
		return;
	if (!(root->right) && !(root->left)) {
		if (!free_function)
			free(root->data);
		else
			free_function(root->data);
		free(root);
		return;
	}
	/* Recursive case. */
	if (root->right) {
		free_tree_data(root->right, free_function);
		root->right = NULL;
	}
	if (root->left) {
		free_tree_data(root->left, free_function);
		root->left = NULL;
	}
	/* Free the current node. */
	if (!free_function)
		free(root->data);
	else
		free_function(root->data);
	free(root);
}

/* Note that the implementation disallows duplication! */
/* This function inserts data into the tree, will fail on invalid input. Comparison function should return 0 for equal, 1 for larger than and -1 for less than. */
static inline void insert_node_into_tree(tree_node_t *root, tree_node_t* const to_insert, int (*comparison_function)(const void*, const void*)) {
	/* Local variables. */
	int comparison_result;

	/* Sanity check. */
	assert(comparison_function && root);

	/* Compare. */
	comparison_result = (*comparison_function)(root->data, to_insert->data);
	/* Check if the value is already there. */
	if (comparison_result == TREE_COMPARISON_EQUAL_TO)
		return;
	/* Check if the root is smaller. */
	if (comparison_result == TREE_COMPARISON_SMALLER_THAN) {
		/* The root is smaller. */
		if (!(root->right)) {
			/* Insert to the right. */
			root->right = to_insert;
			return;
		}
		/* Recurse. */
		insert_node_into_tree(root->right, to_insert, comparison_function);
	} else {
		/* The root is larger. */
		if (!(root->left)) {
			/* Insert to the left. */
			root->left = to_insert;
			return;
		}
		/* Recurse. */
		insert_node_into_tree(root->left, to_insert, comparison_function);
	}
	/* Shouldn't ever reach here! */
}

/* This function finds the tree's depth (at the deepest leaf). */
static inline size_t find_tree_depth(const tree_node_t * const root) {
	/* Simple case. */
	if (!root)
		return ((size_t)0L);

	/* Recurse and return the largest value. */
	return ((size_t)1L) + max(find_tree_depth(root->left), find_tree_depth(root->right));
}
